package net.seesaa.nekokiss.util.dungeon_generator.gui;

import java.awt.Color;
import java.util.List;
import java.util.ArrayList;

import net.seesaa.nekokiss.util.dungeon_generator.Map;
import net.seesaa.nekokiss.util.dungeon_generator.Room;

final class GraphicalMap extends Map {

	private static final List<GraphicalGrid> gridList =
			new ArrayList<GraphicalGrid>();
	private static Color roomColor;
	private static Color passageColor;
	private static Color doorwayColor;
	
	GraphicalMap(int seedX, int seedY, int seedZ, int seedW, int mapWidth,
			int mapHeight, int minRooms, int maxRooms, int minRoomWidth,
			int maxRoomWidth, int minRoomHeight, int maxRoomHeight,
			int maxConnectLength, int tinyRoomRatio, int tinyPassageRatio,
			int intersectRatio, int modPassages, int permissibleFail) {
		super(seedX, seedY, seedZ, seedW, mapWidth, mapHeight, minRooms,
				maxRooms, minRoomWidth, maxRoomWidth, minRoomHeight,
				maxRoomHeight, maxConnectLength, tinyRoomRatio,
				tinyPassageRatio, intersectRatio, modPassages, permissibleFail);
	}
	
	@Override
	protected final void setGrid(int x, int y, byte value) {
		super.setGrid(x, y, value);
		gridList.add(new GraphicalGrid(x, y, value == GRID_ROOM ?
				roomColor : value == GRID_PASSAGE ?
						passageColor : doorwayColor));
	}
	
	@Override
	protected final void setRoom(Room room, boolean tinyRoom,
			boolean tinyPassage) {
		super.setRoom(room, tinyRoom, tinyPassage);
		if (!tinyRoom) {
			GraphicalGrid grid = gridList
					.get(gridList.size() - room.width * room.height);
			grid.type = 1;
			grid.room = room;
			gridList.get(gridList.size() - 1).type = 2;
		}
	}
	
	static final void clearGraphicalGrids() {
		gridList.clear();
	}
	
	static final void setRoomColor(int red, int green, int blue) {
		roomColor = new Color(red, green, blue);
	}
	
	static final void setPassageColor(int red, int green, int blue) {
		passageColor = new Color(red, green, blue);
	}
	
	static final void setDoorwayColor(int red, int green, int blue) {
		doorwayColor = new Color(red, green, blue);
	}
	
	final GraphicalGrid getGrid(int index) {
		return gridList.get(index);
	}
	
	final int getGridsLength() {
		return gridList.size();
	}

}
