package net.seesaa.nekokiss.util.dungeon_generator.gui;

import java.awt.Dimension;
import java.awt.Graphics;
import java.awt.event.ActionListener;
import java.awt.event.ActionEvent;
import java.awt.image.BufferedImage;

import javax.swing.JPanel;
import javax.swing.Timer;

final class GraphicalPanel extends JPanel implements ActionListener {
	
	private final GraphicalFrame parent;
	private final Timer timer;
	private int roomDrawSpeed = 1;
	private int passageDrawSpeed = 15;
	private boolean fillRoom = false;
	private GraphicalMap gMap;
	private BufferedImage bImage;
	private int count;

	GraphicalPanel(GraphicalFrame parent) {
		super(null);
		this.parent = parent;
		this.timer = new Timer(this.roomDrawSpeed, this);
	}

	@Override
	protected final void paintComponent(Graphics g) {
		super.paintComponent(g);
		g.drawImage(this.bImage, 0, 0, null);
	}
	
	@Override
	public final void actionPerformed(ActionEvent e) {
		Graphics g = this.bImage.getGraphics();
		GraphicalGrid graphicalGrid = this.gMap.getGrid(this.count);
		g.setColor(graphicalGrid.color);
		if (graphicalGrid.type > 0) {
			this.timer.setDelay(graphicalGrid.type == 1 ?
					this .roomDrawSpeed : this.passageDrawSpeed);
		}
		if (graphicalGrid.room != null && this.fillRoom) {
			g.fillRect(graphicalGrid.x << 2, graphicalGrid.y << 2, graphicalGrid
					.room.width << 2, graphicalGrid.room.height << 2);
			this.count += graphicalGrid.room.width * graphicalGrid.room
					.height - 1;
			this.timer.setDelay(this.passageDrawSpeed);
		} else {
			g.fillRect(graphicalGrid.x << 2, graphicalGrid.y << 2, 4, 4);
		}
		g.dispose();
		this.repaint();
		if (++this.count >= this.gMap.getGridsLength()) {
			this.timer.stop();
			this.parent.drew(this.gMap);
		}
	}
	
	final void generate(boolean animation, int seedX, int seedY, int seedZ,
			int seedW, int mapWidth, int mapHeight, int minRooms, int maxRooms,
			int minRoomWidth, int maxRoomWidth, int minRoomHeight,
			int maxRoomHeight, int maxConnectRoomsInterval, int tinyRoomRate,
			int tinyPassageRate, int intersectRate, int modPassages,
			int permissibleFail) {
		this.gMap = new GraphicalMap(seedX, seedY, seedZ, seedW, mapWidth,
				mapHeight, minRooms, maxRooms, minRoomWidth, maxRoomWidth,
				minRoomHeight, maxRoomHeight, maxConnectRoomsInterval,
				tinyRoomRate, tinyPassageRate, intersectRate, modPassages,
				permissibleFail);
		this.setPreferredSize(new Dimension(mapWidth << 2, mapHeight << 2));
		this.revalidate();
		this.bImage = new BufferedImage(mapWidth << 2, mapHeight << 2,
				BufferedImage.TYPE_INT_RGB);
		this.count = 0;
		if (animation) {
			this.timer.setDelay(this.roomDrawSpeed);
			this.timer.start();
		} else {
			Graphics g = this.bImage.getGraphics();
			for (int i = 0, length = this.gMap.getGridsLength(); i < length;
					++i) {
				GraphicalGrid graphicalGrid = this.gMap.getGrid(i);
				g.setColor(graphicalGrid.color);
				g.fillRect(graphicalGrid.x << 2, graphicalGrid.y << 2, 4, 4);
			}
			g.dispose();
			this.repaint();
			this.parent.drew(this.gMap);
		}
	}
	
	final void destroy() {
		this.bImage = null;
		this.setVisible(false);
		this.setPreferredSize(new Dimension(0, 0));
		this.revalidate();
	}

	final void setRoomDrawSpeed(int roomDrawSpeed) {
		this.roomDrawSpeed = roomDrawSpeed;
	}

	final void setPassageDrawSpeed(int passageDrawSpeed) {
		this.passageDrawSpeed = passageDrawSpeed;
	}

	final void setFillRoom(boolean fillRoom) {
		this.fillRoom = fillRoom;
	}
	
	final BufferedImage getBufferedImage() {
		return this.bImage;
	}

}
