package net.seesaa.nekokiss.util.dungeon_generator;

import java.util.ArrayList;
import java.util.List;
import java.awt.Point;

import net.seesaa.nekokiss.util.Xor128;

public class Map {

	// 安全のため、各種最大値は小さめにしてある
	public static final byte GRID_WALL = 0;
	public static final byte GRID_PASSAGE = 1;
	public static final byte GRID_ROOM = 2;
	public static final byte GRID_DOORWAY = 3;
	public static final int MIN_MAP_WIDTH = 20;
	public static final int MAX_MAP_WIDTH = 1000;
	public static final int MIN_MAP_HEIGHT = 20;
	public static final int MAX_MAP_HEIGHT = 1000;
	public static final int MIN_ROOMS = 1;
	public static final int MAX_ROOMS = 100;
	public static final int MIN_ROOM_WIDTH = 1;
	public static final int MAX_ROOM_WIDTH = 64;
	public static final int MIN_ROOM_HEIGHT = 1;
	public static final int MAX_ROOM_HEIGHT = 64;
	// この定数名、違和感ある
	public static final int MIN_MAX_CONNECT_LENGTH = 2;
	public static final int MAX_MAX_CONNECT_LENGTH = 128;
	public static final int MIN_TINY_ROOM_RATIO = 0;
	public static final int MAX_TINY_ROOM_RATIO = 100;
	public static final int MIN_TINY_PASSAGE_RATIO = 0;
	public static final int MAX_TINY_PASSAGE_RATIO = 100;
	public static final int MIN_INTERSECT_RATIO = 0;
	public static final int MAX_INTERSECT_RATIO = 100;
	public static final int MIN_MOD_PASSAGES = 0;
	public static final int MAX_MOD_PASSAGES = 100;
	public static final int MIN_PERMISSIBLE_FAILS = 0;
	public static final int MAX_PERMISSIBLE_FAILS = 20000;
	private static final Xor128 RND = new Xor128();
	private final int seedX;
	private final int seedY;
	private final int seedZ;
	private final int seedW;
	private final int mapWidth;
	private final int mapHeight;
	private final int minRooms;
	private final int maxRooms;
	private final int minRoomWidth;
	private final int maxRoomWidth;
	private final int minRoomHeight;
	private final int maxRoomHeight;
	private final int maxConnectLength;
	private final int tinyRoomRatio;
	private final int tinyPassageRatio;
	private final int intersectRatio;
	private final int modPassages;
	private final int permissibleFails;
	private final byte[] grids;
	private final Room[] rooms;
	private final int fails;

	public Map(int seedX, int seedY, int seedZ, int seedW, int mapWidth,
			int mapHeight, int minRooms, int maxRooms, int minRoomWidth,
			int maxRoomWidth, int minRoomHeight, int maxRoomHeight,
			int maxConnectLength, int tinyRoomRatio, int tinyPassageRatio,
			int intersectRatio, int modPassages, int permissibleFails) {
		if (mapWidth < MIN_MAP_WIDTH) {
			throw new IllegalArgumentException("mapWidth は、" + MIN_MAP_WIDTH +
					" 以上である必要があります。");
		}
		if (mapWidth > MAX_MAP_WIDTH) {
			throw new IllegalArgumentException("mapWidth は、" + MAX_MAP_WIDTH +
					" 以下である必要があります。");
		}
		if (mapHeight < MIN_MAP_HEIGHT) {
			throw new IllegalArgumentException("mapHeight は、" +
					MIN_MAP_HEIGHT + " 以上である必要があります。");
		}
		if (mapHeight > MAX_MAP_HEIGHT) {
			throw new IllegalArgumentException("mapHeight は、" +
					MAX_MAP_HEIGHT + " 以下である必要があります。");
		}
		if (minRooms < MIN_ROOMS) {
			throw new IllegalArgumentException("minRooms は、" + MIN_ROOMS +
					" 以上である必要があります。");
		}
		if (minRooms > MAX_ROOMS) {
			throw new IllegalArgumentException("minRooms は、" + MAX_ROOMS +
					" 以下である必要があります。");
		}
		if (maxRooms < MIN_ROOMS) {
			throw new IllegalArgumentException("maxRooms は、" + MIN_ROOMS +
					" 以上である必要があります。");
		}
		if (maxRooms > MAX_ROOMS) {
			throw new IllegalArgumentException("maxRooms は、" + MAX_ROOMS +
					" 以下である必要があります。");
		}
		if (minRoomWidth < MIN_ROOM_WIDTH) {
			throw new IllegalArgumentException("minRoomWidth は、" +
					MIN_ROOM_WIDTH + " 以上である必要があります。");
		}
		if (minRoomWidth > MAX_ROOM_WIDTH) {
			throw new IllegalArgumentException("minRoomWidth は、" +
					MAX_ROOM_WIDTH + " 以下である必要があります。");
		}
		if (maxRoomWidth < MIN_ROOM_WIDTH) {
			throw new IllegalArgumentException("maxRoomWidth は、" +
					MIN_ROOM_WIDTH + " 以上である必要があります。");
		}
		if (maxRoomWidth > MAX_ROOM_WIDTH) {
			throw new IllegalArgumentException("maxRoomWidth は、" +
					MAX_ROOM_WIDTH + " 以下である必要があります。");
		}
		if (minRoomHeight < MIN_ROOM_HEIGHT) {
			throw new IllegalArgumentException("minRoomHeight は、" +
					MIN_ROOM_HEIGHT + " 以上である必要があります。");
		}
		if (minRoomHeight > MAX_ROOM_HEIGHT) {
			throw new IllegalArgumentException("minRoomHeight は、" +
					MAX_ROOM_HEIGHT + " 以下である必要があります。");
		}
		if (maxRoomHeight < MIN_ROOM_HEIGHT) {
			throw new IllegalArgumentException("maxRoomHeight は、" +
					MIN_ROOM_HEIGHT + " 以上である必要があります。");
		}
		if (maxRoomHeight > MAX_ROOM_HEIGHT) {
			throw new IllegalArgumentException("maxRoomHeight は、" +
					MAX_ROOM_HEIGHT + " 以下である必要があります。");
		}
		if (maxConnectLength < MIN_MAX_CONNECT_LENGTH) {
			throw new IllegalArgumentException("maxConnectRoomsInterval は、" +
					MIN_MAX_CONNECT_LENGTH +
					" 以上である必要があります。");
		}
		if (maxConnectLength > MAX_MAX_CONNECT_LENGTH) {
			throw new IllegalArgumentException("maxConnectRoomsInterval は、" +
					MAX_MAX_CONNECT_LENGTH +
					" 以下である必要があります。");
		}
		if (tinyRoomRatio < MIN_TINY_ROOM_RATIO) {
			throw new IllegalArgumentException("tinyRoomRatio は、" +
					MIN_TINY_ROOM_RATIO + " 以上である必要があります。");
		}
		if (tinyRoomRatio > MAX_TINY_ROOM_RATIO) {
			throw new IllegalArgumentException("tinyRoomRatio は、" +
					MAX_TINY_ROOM_RATIO + " 以下である必要があります。");
		}
		if (tinyPassageRatio < MIN_TINY_PASSAGE_RATIO) {
			throw new IllegalArgumentException("tinyPassageRatio は、" +
					MIN_TINY_PASSAGE_RATIO + " 以上である必要があります。");
		}
		if (tinyPassageRatio > MAX_TINY_PASSAGE_RATIO) {
			throw new IllegalArgumentException("tinyPassageRatio は、" +
					MAX_TINY_PASSAGE_RATIO + " 以下である必要があります。");
		}
		if (intersectRatio < MIN_INTERSECT_RATIO) {
			throw new IllegalArgumentException("intersectRatio は、" +
					MIN_INTERSECT_RATIO + " 以上である必要があります。");
		}
		if (intersectRatio > MAX_INTERSECT_RATIO) {
			throw new IllegalArgumentException("intersectRatio は、" +
					MAX_INTERSECT_RATIO + " 以下である必要があります。");
		}
		if (modPassages < MIN_MOD_PASSAGES) {
			throw new IllegalArgumentException("modPassages は、" +
					MIN_MOD_PASSAGES + " 以上である必要があります。");
		}
		if (modPassages > MAX_MOD_PASSAGES) {
			throw new IllegalArgumentException("modPassages は、" +
					MAX_MOD_PASSAGES + " 以下である必要があります。");
		}
		if (permissibleFails < MIN_PERMISSIBLE_FAILS) {
			throw new IllegalArgumentException("permissibleFails は、" +
					MIN_PERMISSIBLE_FAILS + " 以上である必要があります。");
		}
		if (permissibleFails > MAX_PERMISSIBLE_FAILS) {
			throw new IllegalArgumentException("permissibleFails は、" +
					MAX_PERMISSIBLE_FAILS + " 以下である必要があります。");
		}
		if (maxRooms < minRooms) {
			throw new IllegalArgumentException("maxRooms は、minRooms" +
					" 以上である必要があります。");
		}
		if (minRoomWidth > mapWidth - 2) {
			throw new IllegalArgumentException("minRoomWidth は、mapWidth - 2" +
					" 以下である必要があります。");
		}
		if (maxRoomWidth > mapWidth - 2) {
			throw new IllegalArgumentException("maxRoomWidth は、mapWidth - 2" +
					" 以下である必要があります。");
		}
		if (minRoomHeight > mapHeight - 2) {
			throw new IllegalArgumentException("minRoomHeight は、mapHeight -" +
					" 2 以下である必要があります。");
		}
		if (maxRoomHeight > mapHeight - 2) {
			throw new IllegalArgumentException("maxRoomHeight は、mapHeight -" +
					" 2 以下である必要があります。");
		}
		if (maxRoomWidth < minRoomWidth) {
			throw new IllegalArgumentException("maxRoomWidth は、minRoomWidth" +
					" 以上である必要があります。");
		}
		if (maxRoomHeight < minRoomHeight) {
			throw new IllegalArgumentException("maxRoomHeight は、" +
					"minRoomHeight 以上である必要があります。");
		}
		if ((seedX | seedY | seedZ | seedW) != 0) {
			RND.setSeeds(seedX, seedY, seedZ, seedW);
			this.seedX = seedX;
			this.seedY = seedY;
			this.seedZ = seedZ;
			this.seedW = seedW;
		} else {
			if (RND.isEmpty()) {
				RND.setSeeds();
			}
			this.seedX = RND.getX();
			this.seedY = RND.getY();
			this.seedZ = RND.getZ();
			this.seedW = RND.getW();
		}
		this.mapWidth = mapWidth;
		this.mapHeight = mapHeight;
		this.minRooms = minRooms;
		this.maxRooms = maxRooms;
		this.minRoomWidth = minRoomWidth;
		this.maxRoomWidth = maxRoomWidth;
		this.minRoomHeight = minRoomHeight;
		this.maxRoomHeight = maxRoomHeight;
		this.maxConnectLength = maxConnectLength;
		this.tinyRoomRatio = tinyRoomRatio;
		this.tinyPassageRatio = tinyPassageRatio;
		this.intersectRatio = intersectRatio;
		this.modPassages = modPassages;
		this.permissibleFails = permissibleFails;
		this.grids = new byte[mapWidth * mapHeight];
		List<Room> roomList = new ArrayList<Room>();
		int fails = 0;
		boolean fullFails = false;
		start: for (int i = 0, rooms = this.minRooms < this.maxRooms ?
				RND.next(this.minRooms, this.maxRooms + 1) : this.minRooms,
				width, height; i < rooms;) {
			boolean tinyRoom, tinyPassage = false;
			if (this.tinyRoomRatio >= 100 || (this.tinyRoomRatio > 0 &&
					RND.next(100) < this.tinyRoomRatio)) {
				width = height = 1;
				tinyRoom = true;
				if (tinyRoom && (this.tinyPassageRatio >= 100 ||
						(this.tinyPassageRatio > 0 &&
								RND.next(100) < this.tinyPassageRatio))) {
					tinyPassage = true;
				}
			} else {
				width = RND.next(this.minRoomWidth, this.maxRoomWidth + 1);
				height = RND.next(this.minRoomWidth, this.maxRoomWidth + 1);
				tinyRoom = false;
			}
			// 座標外の判定が面倒なので、マップの四隅に部屋が重ならないように
			Room room = new Room(RND.next(this.mapWidth - width - 1) + 1,
					RND.next(this.mapHeight - height - 1) + 1, width, height,
					tinyPassage);
			for (int y = room.y - 1; y <= room.bottom + 1; ++y) {
				for (int x = room.x - 1; x <= room.right + 1; ++x) {
					if (this.getGrid(x, y) != GRID_WALL) {
						if (++fails > this.permissibleFails) {
							fullFails = true;
							break start;
						}
						continue start;
					}
				}
			}
			if (i == 0) {
				this.setRoom(room, tinyRoom, tinyPassage);
				roomList.add(room);
				++i;
				continue;
			}
			Point[] path;
			for (int j = 0, index = i == 1 ? 0 : RND.next(i);;) {
				path = this.generatePath(room, roomList.get(index));
				if (path.length > 0) {
					this.setRoom(room, tinyRoom, tinyPassage);
					// Debug
					/*System.out.println(room.x + ", " + room.y + ", " + room.width + ", " + room.height);
					for (int dbg = 0; ;) {
						System.out.print(path[dbg]);
						if (++dbg < path.length) {
							System.out.print(", ");
						} else {
							System.out.println();
							break;
						}
					}*/
					break;
				} else if (++j == i) {
					if (++fails > this.permissibleFails) {
						fullFails = true;
						break start;
					}
					continue start;
				} else if (++index == i) {
					index = 0;
				}
			}
			this.setPath(path, tinyPassage);
			roomList.add(room);
			++i;
		}
		if (!fullFails && this.modPassages > 0 && roomList.size() > 1) {
			for (int i = 0, length = roomList.size(), index1, index2;
					i < this.modPassages;) {
				index1 = RND.next(length);
				while (true) {
					index2 = RND.next(length);
					if (index1 != index2) {
						break;
					}
				}
				Room room1 = roomList.get(index1);
				Point[] path = this.generatePath(room1, roomList.get(index2));
				if (path.length > 0) {
					this.setPath(path, room1.passage);
					++i;
				} else {
					if (++fails > this.permissibleFails) {
						fullFails = true;
						break;
					}
				}
			}
		}
		this.rooms = (Room[])roomList.toArray(new Room[roomList.size()]);
		this.fails = fails;
	}

	private final Point[] generatePath(Room room1, Room room2) {
		List<Point> pointList = new ArrayList<Point>();
		outside: for (int i = 0, side = RND.next(4); i < 4; ++i) {
			switch (side) {
			case 0: // Top
				if (room1.y <= room2.y || room1.y - room2.bottom >
						this.maxConnectLength) {
					break;
				}
				for (int j = 0, direction = RND.next(3); j < 3; ++j) {
					switch (direction) {
					case 0: // North West
						if (this.generateNorthWestPath(room1, room2,
								pointList)) {
							break outside;
						}
						break;
					case 1: // North
						if (this.generateNorthPath(room1, room2, pointList)) {
							break outside;
						}
						break;
					case 2: // North East
						if (this.generateNorthEastPath(room1, room2,
								pointList)) {
							break outside;
						}
						break;
					}
					if (++direction == 3) {
						direction = 0;
					}
				}
				break;
			case 1: // Bottom
				if (room2.bottom <= room1.bottom || room2.y - room1.bottom >
						this.maxConnectLength) {
					break;
				}
				for (int l = 0, direction = RND.next(3); l < 3; ++l) {
					switch (direction) {
					case 0: // South West
						if (this.generateSouthWestPath(room1, room2,
								pointList)) {
							break outside;
						}
						break;
					case 1: // South
						if (this.generateSouthPath(room1, room2, pointList)) {
							break outside;
						}
						break;
					case 2: // South East
						if (this.generateSouthEastPath(room1, room2,
								pointList)) {
							break outside;
						}
						break;
					}
					if (++direction == 3) {
						direction = 0;
					}
				}
				break;
			case 2: // Left
				if (room1.x <= room2.x || room1.x - room2.right >
						this.maxConnectLength) {
					break;
				}
				for (int l = 0, direction = RND.next(3); l < 3; ++l) {
					switch (direction) {
					case 0: // West North
						if (this.generateWestNorthPath(room1, room2,
								pointList)) {
							break outside;
						}
						break;
					case 1: // West
						if (this.generateWestPath(room1, room2, pointList)) {
							break outside;
						}
						break;
					case 2: // West South
						if (this.generateWestSouthPath(room1, room2,
								pointList)) {
							break outside;
						}
						break;
					}
					if (++direction == 3) {
						direction = 0;
					}
				}
				break;
			case 3: // Right
				if (room2.right <= room1.right || room2.x - room1.right >
						this.maxConnectLength) {
					break;
				}
				for (int l = 0, direction = RND.next(3); l < 3; ++l) {
					switch (direction) {
					case 0: // East North
						if (this.generateEastNorthPath(room1, room2,
								pointList)) {
							break outside;
						}
						break;
					case 1: // East
						if (this.generateEastPath(room1, room2, pointList)) {
							break outside;
						}
						break;
					case 2: // East South
						if (this.generateEastSouthPath(room1, room2,
								pointList)) {
							break outside;
						}
						break;
					}
					if (++direction == 3) {
						direction = 0;
					}
				}
				break;
			}
			if (++side == 4) {
				side = 0;
			}
		}
		return pointList.toArray(new Point[pointList.size()]);
	}

	private final boolean generateNorthWestPath(Room room1, Room room2,
			List<Point> pointList) {
		if (room1.right <= room2.right
				|| room1.x - room2.right > this.maxConnectLength) {
			return false;
		}
		pointList.add(new Point(RND.next(Math.max(room1.x, room2.x),
				room1.right + 1), room1.y - 1));
		if (pointList.get(0).x > room2.right) {
			int y2 = RND.next(room2.y, Math.min(room2.bottom + 1, room1.y));
			if (y2 == pointList.get(0).y) {
				pointList.clear();
				return false;
			}
			pointList.add(new Point(pointList.get(0).x, y2));
			pointList.add(new Point(room2.right + 1, pointList.get(1).y));
			return true;
		}
		pointList.add(new Point(pointList.get(0).x, room2.bottom + 1));
		return true;
	}

	private final boolean generateNorthPath(Room room1, Room room2,
			List<Point> pointList) {
		if ((room1.x > room2.x && room1.right > room2.right)
				|| (room2.x > room1.x && room2.right > room1.right)) {
			return false;
		}
		pointList.add(new Point(RND.next(Math.max(room1.x, room2.x),
				Math.min(room1.right, room2.right) + 1), room1.y - 1));
		pointList.add(new Point(pointList.get(0).x, room2.bottom + 1));
		return true;
	}

	private final boolean generateNorthEastPath(Room room1, Room room2,
			List<Point> pointList) {
		if (room2.x <= room1.x
				|| room2.x - room1.right > this.maxConnectLength) {
			return false;
		}
		pointList.add(new Point(RND.next(room1.x,
				Math.min(room1.right, room2.right) + 1), room1.y - 1));
		if (room2.x > pointList.get(0).x) {
			int y2 = RND.next(room2.y, Math.min(room2.bottom + 1, room1.y));
			if (y2 == pointList.get(0).y) {
				pointList.clear();
				return false;
			}
			pointList.add(new Point(pointList.get(0).x, y2));
			pointList.add(new Point(room2.x - 1, pointList.get(1).y));
			return true;
		}
		pointList.add(new Point(pointList.get(0).x, room2.bottom + 1));
		return true;
	}

	private final boolean generateSouthWestPath(Room room1, Room room2,
			List<Point> pointList) {
		if (room1.right <= room2.right
				|| room1.x - room2.right > this.maxConnectLength) {
			return false;
		}
		pointList.add(new Point(RND.next(Math.max(room1.x, room2.x),
				room1.right + 1), room1.bottom + 1));
		if (pointList.get(0).x > room2.right) {
			int y2 = RND
					.next(Math.max(room2.y, room1.bottom + 1), room2.bottom);
			if (y2 == pointList.get(0).y) {
				pointList.clear();
				return false;
			}
			pointList.add(new Point(pointList.get(0).x, y2));
			pointList.add(new Point(room2.right + 1, pointList.get(1).y));
			return true;
		}
		pointList.add(new Point(pointList.get(0).x, room2.y - 1));
		return true;
	}

	private final boolean generateSouthPath(Room room1, Room room2,
			List<Point> pointList) {
		if ((room1.x > room2.x && room1.right > room2.right)
				|| (room2.x > room1.x && room2.right > room1.right)) {
			return false;
		}
		pointList.add(new Point(RND.next(Math.max(room1.x, room2.x),
				Math.min(room1.right, room2.right) + 1), room1.bottom + 1));
		pointList.add(new Point(pointList.get(0).x, room2.y - 1));
		return true;
	}

	private final boolean generateSouthEastPath(Room room1, Room room2,
			List<Point> pointList) {
		if (room2.x <= room1.x
				|| room2.x - room1.right > this.maxConnectLength) {
			return false;
		}
		pointList.add(new Point(RND.next(room1.x,
				Math.min(room1.right, room2.right) + 1), room1.bottom + 1));
		if (room2.x > pointList.get(0).x) {
			int y2 = RND.next(Math.max(room2.y, room1.bottom + 1),
					room2.bottom + 1);
			if (y2 == pointList.get(0).y) {
				pointList.clear();
				return false;
			}
			pointList.add(new Point(pointList.get(0).x, y2));
			pointList.add(new Point(room2.x - 1, pointList.get(1).y));
			return true;
		}
		pointList.add(new Point(pointList.get(0).x, room2.y - 1));
		return true;
	}

	private final boolean generateWestNorthPath(Room room1, Room room2,
			List<Point> pointList) {
		if (room1.bottom <= room2.bottom
				|| room1.y - room2.bottom > this.maxConnectLength) {
			return false;
		}
		pointList.add(new Point(room1.x - 1, RND.next(
				Math.max(room1.y, room2.y), room1.bottom + 1)));
		if (pointList.get(0).y > room2.bottom) {
			int x2 = RND.next(room2.x, Math.min(room2.right + 1, room1.x));
			if (x2 == pointList.get(0).x) {
				pointList.clear();
				return false;
			}
			pointList.add(new Point(x2, pointList.get(0).y));
			pointList.add(new Point(pointList.get(1).x, room2.bottom + 1));
			return true;
		}
		pointList.add(new Point(room2.right + 1, pointList.get(0).y));
		return true;
	}

	private final boolean generateWestPath(Room room1, Room room2,
			List<Point> pointList) {
		if ((room1.y > room2.y && room1.bottom > room2.bottom)
				|| (room2.y > room1.y && room2.bottom > room1.bottom)) {
			return false;
		}
		pointList.add(new Point(room1.x - 1, RND.next(
				Math.max(room1.y, room2.y),
				Math.min(room1.bottom, room2.bottom) + 1)));
		pointList.add(new Point(room2.right + 1, pointList.get(0).y));
		return true;
	}

	private final boolean generateWestSouthPath(Room room1, Room room2,
			List<Point> pointList) {
		if (room2.y <= room1.y
				|| room2.y - room1.bottom > this.maxConnectLength) {
			return false;
		}
		pointList.add(new Point(room1.x - 1, RND.next(room1.y,
				Math.min(room1.bottom, room2.bottom) + 1)));
		if (room2.y > pointList.get(0).y) {
			int x2 = RND.next(room2.x, Math.min(room2.right + 1, room1.x));
			if (x2 == pointList.get(0).x) {
				pointList.clear();
				return false;
			}
			pointList.add(new Point(x2, pointList.get(0).y));
			pointList.add(new Point(pointList.get(1).x, room2.y - 1));
			return true;
		}
		pointList.add(new Point(room2.right + 1, pointList.get(0).y));
		return true;
	}

	private final boolean generateEastNorthPath(Room room1, Room room2,
			List<Point> pointList) {
		if (room1.bottom <= room2.bottom
				|| room1.y - room2.bottom > this.maxConnectLength) {
			return false;
		}
		pointList.add(new Point(room1.right + 1, RND.next(
				Math.max(room1.y, room2.y), room1.bottom + 1)));
		if (pointList.get(0).y > room2.bottom) {
			int x2 = RND.next(Math.max(room2.x, room1.right + 1), room2.right);
			if (x2 == pointList.get(0).x) {
				pointList.clear();
				return false;
			}
			pointList.add(new Point(x2, pointList.get(0).y));
			pointList.add(new Point(pointList.get(1).x, room2.bottom + 1));
			return true;
		}
		pointList.add(new Point(room2.right - 1, pointList.get(0).y));
		return true;
	}

	private final boolean generateEastPath(Room room1, Room room2,
			List<Point> pointList) {
		if ((room1.y > room2.y && room1.bottom > room2.bottom)
				|| (room2.y > room1.y && room2.bottom > room1.bottom)) {
			return false;
		}
		pointList.add(new Point(room1.right + 1, RND.next(
				Math.max(room1.y, room2.y),
				Math.min(room1.bottom, room2.bottom) + 1)));
		pointList.add(new Point(room2.x - 1, pointList.get(0).y));
		return true;
	}

	private final boolean generateEastSouthPath(Room room1, Room room2,
			List<Point> pointList) {
		if (room2.y <= room1.y
				|| room2.y - room1.bottom > this.maxConnectLength) {
			return false;
		}
		pointList.add(new Point(room1.right + 1, RND.next(room1.y,
				Math.min(room1.bottom, room2.bottom) + 1)));
		if (room2.y > pointList.get(0).y) {
			int x2 = RND.next(Math.max(room2.x, room1.right + 1),
					room2.right + 1);
			if (x2 == pointList.get(0).x) {
				pointList.clear();
				return false;
			}
			pointList.add(new Point(x2, pointList.get(0).y));
			pointList.add(new Point(pointList.get(1).x, room2.y - 1));
			return true;
		}
		pointList.add(new Point(room2.right - 1, pointList.get(0).y));
		return true;
	}

	private final void setPath(Point[] path, boolean tinyPassage) {
		outside: for (int i = 1; i < path.length; ++i) {
			int x = path[i - 1].x, y = path[i - 1].y,
					x2 = path[i].x, y2 = path[i].y;
			if (x == x2) {
				if (y == y2) {
					this.setGrid(x, y, (this.getGrid(x - 1, y) == GRID_ROOM ||
							this.getGrid(x + 1, y) == GRID_ROOM ||
							this.getGrid(x, y - 1) == GRID_ROOM ||
							this.getGrid(x, y + 1) == GRID_ROOM) ?
									GRID_DOORWAY : GRID_PASSAGE);
					continue;
				}
				if (y > y2) {
					// ↑
					if (i == 1) {
						if (this.getGrid(x - 1, y) != GRID_WALL
								|| this.getGrid(x + 1, y) != GRID_WALL) {
							if (this.getGrid(x, y - 1) == GRID_ROOM ||
									this.getGrid(x, y + 1) == GRID_ROOM) {
								this.setGrid(x, y, tinyPassage ? GRID_PASSAGE
										: GRID_DOORWAY);
							}
							break;
						}
						this.setGrid(x, y, tinyPassage ? GRID_PASSAGE
								: GRID_DOORWAY);
						if (this.getGrid(x, y - 1) != GRID_WALL) {
							break;
						}
					} else {
						this.setGrid(x, y, GRID_PASSAGE);
						if ((this.getGrid(x - 1, y) != GRID_WALL &&
								this.getGrid(x - 1, y - 1) != GRID_WALL) ||
								(this.getGrid(x + 1, y) != GRID_WALL &&
								this.getGrid(x + 1, y - 1) != GRID_WALL)) {
							break;
						}
					}
					while (--y >= y2) {
						if (this.getGrid(x, y - 1) == GRID_ROOM
								|| getGrid(x - 1, y) == GRID_ROOM
								|| getGrid(x + 1, y) == GRID_ROOM) {
							this.setGrid(x, y, GRID_DOORWAY);
							break outside;
						} else if (this.getGrid(x, y) != GRID_WALL) {
							if (this.getGrid(x, y - 1) != GRID_WALL
									|| (this.intersectRatio < 100 && RND
											.next(100) >= this
											.intersectRatio)) {
								break outside;
							}
							if (y - 1 >= y2) {
								break outside;
							}
						}
						if (y == y2 && i == 1) {
							if (this.getGrid(x, y - 1) != GRID_WALL
									|| getGrid(x - 1, y) != GRID_WALL
									|| getGrid(x + 1, y) != GRID_WALL) {
								this.setGrid(x, y, GRID_PASSAGE);
							}
							break;
						}
						this.setGrid(x, y, GRID_PASSAGE);
						if (this.getGrid(x - 1, y) != GRID_WALL) {
							if (this.getGrid(x - 1, y - 1) != GRID_WALL) {
								if (y > y2 && i == 1) {
									this.setGrid(x, y, GRID_PASSAGE);
								}
								break outside;
							}
						}
						if (this.getGrid(x + 1, y) != GRID_WALL) {
							if (this.getGrid(x + 1, y - 1) != GRID_WALL) {
								if (y > y2 && i == 1) {
									this.setGrid(x, y, GRID_PASSAGE);
								}
								break outside;
							}
						}
					}
				} else {
					// ↓
					if (i == 1) {
						if (this.getGrid(x - 1, y) != GRID_WALL
								|| getGrid(x + 1, y) != GRID_WALL) {
							if (this.getGrid(x, y + 1) == GRID_ROOM ||
									this.getGrid(x, y - 1) == GRID_ROOM) {
								this.setGrid(x, y, tinyPassage ? GRID_PASSAGE
										: GRID_DOORWAY);
							}
							break;
						}
						this.setGrid(x, y, tinyPassage ? GRID_PASSAGE
								: GRID_DOORWAY);
						if (this.getGrid(x, y + 1) != GRID_WALL) {
							break;
						}
					} else {
						this.setGrid(x, y, GRID_PASSAGE);
						if ((this.getGrid(x - 1, y) != GRID_WALL &&
								this.getGrid(x - 1, y + 1) != GRID_WALL) ||
								(this.getGrid(x + 1, y) != GRID_WALL &&
								this.getGrid(x + 1, y + 1) != GRID_WALL)) {
							break;
						}
					}
					while (++y <= y2) {
						if (this.getGrid(x, y + 1) == GRID_ROOM
								|| this.getGrid(x - 1, y) == GRID_ROOM
								|| this.getGrid(x + 1, y) == GRID_ROOM) {
							this.setGrid(x, y, GRID_DOORWAY);
							break outside;
						} else if (this.getGrid(x, y) != GRID_WALL) {
							if (this.getGrid(x, y + 1) != GRID_WALL
									|| (this.intersectRatio < 100 && RND
											.next(100) >= this
											.intersectRatio)) {
								break outside;
							}
							if (y + 1 <= y2) {
								break outside;
							}
						}
						if (y == y2 && i == 1) {
							if (this.getGrid(x, y + 1) != GRID_WALL
									|| this.getGrid(x - 1, y) != GRID_WALL
									|| this.getGrid(x + 1, y) != GRID_WALL) {
								this.setGrid(x, y, GRID_PASSAGE);
							}
							break;
						}
						this.setGrid(x, y, GRID_PASSAGE);
						if (this.getGrid(x - 1, y) != GRID_WALL) {
							if (this.getGrid(x - 1, y + 1) != GRID_WALL) {
								if (y < y2 && i == 1) {
									this.setGrid(x, y, GRID_PASSAGE);
								}
								break outside;
							}
						}
						if (this.getGrid(x + 1, y) != GRID_WALL) {
							if (this.getGrid(x + 1, y + 1) != GRID_WALL) {
								if (y < y2 && i == 1) {
									this.setGrid(x, y, GRID_PASSAGE);
								}
								break outside;
							}
						}
					}
				}
			} else {
				if (x > x2) {
					// ←
					if (i == 1) {
						if (this.getGrid(x, y - 1) != GRID_WALL
								|| this.getGrid(x, y + 1) != GRID_WALL) {
							if (this.getGrid(x - 1, y) == GRID_ROOM ||
									this.getGrid(x + 1, y) == GRID_ROOM) {
								this.setGrid(x, y, tinyPassage ? GRID_PASSAGE
										: GRID_DOORWAY);
							}
							break;
						}
						this.setGrid(x, y, tinyPassage ? GRID_PASSAGE
								: GRID_DOORWAY);
						if (this.getGrid(x - 1, y) != GRID_WALL) {
							break;
						}
					} else {
						this.setGrid(x, y, GRID_PASSAGE);
						if ((this.getGrid(x, y - 1) != GRID_WALL &&
								this.getGrid(x - 1, y - 1) != GRID_WALL) ||
								(this.getGrid(x, y + 1) != GRID_WALL &&
								this.getGrid(x - 1, y + 1) != GRID_WALL)) {
							break;
						}
					}
					while (--x >= x2) {
						if (this.getGrid(x - 1, y) == GRID_ROOM
								|| this.getGrid(x, y - 1) == GRID_ROOM
								|| this.getGrid(x, y + 1) == GRID_ROOM) {
							this.setGrid(x, y, GRID_DOORWAY);
							break outside;
						} else if (this.getGrid(x, y) != GRID_WALL) {
							if (this.getGrid(x - 1, y) != GRID_WALL
									|| (this.intersectRatio < 100 && RND
											.next(100) >= this
											.intersectRatio)) {
								break outside;
							}
							if (x - 1 >= x2) {
								break outside;
							}
						}
						if (x == x2 && i == 1) {
							if (this.getGrid(x - 1, y) != GRID_WALL
									|| this.getGrid(x, y - 1) != GRID_WALL
									|| this.getGrid(x, y + 1) != GRID_WALL) {
								this.setGrid(x, y, GRID_PASSAGE);
							}
							break;
						}
						this.setGrid(x, y, GRID_PASSAGE);
						if (this.getGrid(x, y - 1) != GRID_WALL) {
							if (this.getGrid(x - 1, y - 1) != GRID_WALL) {
								break outside;
							}
						}
						if (this.getGrid(x, y + 1) != GRID_WALL) {
							if (this.getGrid(x - 1, y + 1) != GRID_WALL) {
								break outside;
							}
						}
					}
				} else {
					// →
					if (i == 1) {
						if (this.getGrid(x, y - 1) != GRID_WALL
								|| this.getGrid(x, y + 1) != GRID_WALL) {
							if (this.getGrid(x + 1, y) == GRID_ROOM ||
									this.getGrid(x - 1, y) == GRID_ROOM) {
								this.setGrid(x, y, tinyPassage ? GRID_PASSAGE
										: GRID_DOORWAY);
							}
							break;
						}
						this.setGrid(x, y, tinyPassage ? GRID_PASSAGE
								: GRID_DOORWAY);
						if (this.getGrid(x + 1, y) != GRID_WALL) {
							break;
						}
					} else {
						this.setGrid(x, y, GRID_PASSAGE);
						if ((this.getGrid(x, y - 1) != GRID_WALL &&
								this.getGrid(x + 1, y - 1) != GRID_WALL) ||
								(this.getGrid(x, y + 1) != GRID_WALL &&
								this.getGrid(x + 1, y + 1) != GRID_WALL)) {
							break;
						}
					}
					while (++x <= x2) {
						if (this.getGrid(x + 1, y) == GRID_ROOM
								|| this.getGrid(x, y - 1) == GRID_ROOM
								|| this.getGrid(x, y + 1) == GRID_ROOM) {
							this.setGrid(x, y, GRID_DOORWAY);
							break outside;
						} else if (this.getGrid(x, y) != GRID_WALL) {
							if (this.getGrid(x + 1, y) != GRID_WALL
									|| (this.intersectRatio < 100 && RND
											.next(100) >= this
											.intersectRatio)) {
								break outside;
							}
							if (x + 1 <= x2) {
								break outside;
							}
						}
						if (x == x2 && i == 1) {
							if (this.getGrid(x + 1, y) != GRID_WALL
									|| this.getGrid(x, y - 1) != GRID_WALL
									|| this.getGrid(x, y + 1) != GRID_WALL) {
								this.setGrid(x, y, GRID_PASSAGE);
							}
							break;
						}
						this.setGrid(x, y, GRID_PASSAGE);
						if (this.getGrid(x, y - 1) != GRID_WALL) {
							if (this.getGrid(x + 1, y - 1) != GRID_WALL) {
								break outside;
							}
						}
						if (this.getGrid(x, y + 1) != GRID_WALL) {
							if (this.getGrid(x + 1, y + 1) != GRID_WALL) {
								break outside;
							}
						}
					}
				}
			}
		}
	}
	
	public final long getSeedX() {
		return this.seedX & 0xffffffffL;
	}

	public final long getSeedY() {
		return this.seedY & 0xffffffffL;
	}

	public final long getSeedZ() {
		return this.seedZ & 0xffffffffL;
	}

	public final long getSeedW() {
		return this.seedW & 0xffffffffL;
	}

	public final int getMapWidth() {
		return this.mapWidth;
	}

	public final int getMapHeight() {
		return this.mapHeight;
	}

	public final int getMinRooms() {
		return this.minRooms;
	}

	public final int getMaxRooms() {
		return this.maxRooms;
	}

	public final int getMinRoomWidth() {
		return this.minRoomWidth;
	}

	public final int getMaxRoomWidth() {
		return this.maxRoomWidth;
	}

	public final int getMinRoomHeight() {
		return this.minRoomHeight;
	}

	public final int getMaxRoomHeight() {
		return this.maxRoomHeight;
	}

	public final int getMaxConnectLength() {
		return this.maxConnectLength;
	}

	public final int getTinyRoomRatio() {
		return this.tinyRoomRatio;
	}

	public final int getTinyPassageRatio() {
		return this.tinyPassageRatio;
	}

	public final int getIntersectRatio() {
		return this.intersectRatio;
	}

	public final int getModPassages() {
		return this.modPassages;
	}

	public final int getPermissibleFails() {
		return this.permissibleFails;
	}
	
	public final int getRoomCount() {
		return this.rooms.length;
	}

	public final Room getRoom(int index) {
		return this.rooms[index];
	}

	public final int getFails() {
		return this.fails;
	}

	public final byte getGrid(int x, int y) {
		return this.grids[x + y * this.mapWidth];
	}

	protected void setGrid(int x, int y, byte value) {
		this.grids[x + y * this.mapWidth] = value;
	}

	protected void setRoom(Room room, boolean tinyRoom, boolean tinyPassage) {
		if (tinyRoom) {
			this.setGrid(room.x, room.y,
					tinyPassage ? GRID_PASSAGE : GRID_ROOM);
		} else {
			for (int y = room.y, bottom = y + room.height; y < bottom; ++y) {
				for (int x = room.x, right = x + room.width; x < right; ++x) {
					this.setGrid(x, y, GRID_ROOM);
				}
			}
		}
	}

}
